# Copyright (C) 2019 Cumulus Networks, Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import struct

from portwd_utils import SFFCable, set_logger

ETH_QSFP_CODE_MAP = {
                           '100g-sr4': ['100G Short Range (SR4) fiber QSFP28 module','50G Short Range (SR) fiber QSFP28 module'],
                           '100g-lr4': ['100G Long Range (LR)4 Single mode fiber QSFP28 module'],
                           '100g-er4': ['100G Extended Range (ER4) Single mode fiber QSFP28 module'],
                           '100g-cwdm4fec': ['100G CWDM4 MSA Group Coarse Wavelength Division Multiplexing (2km) Single mode fiber QSFP28 module'],
                           '100g-psm4': ['100G Parallel Single Mode 4-channel (PSM4) Single mode fiber QSFP28 module'],
                           '100g-active-cu': ['100G Active Copper Direct Attach Copper (DAC) cable QSFP28 to QSFP28 Twinax w/ amplifier and equalization'],
                           '100g-cwdm4': ['100G CWDM4 MSA Group Coarse Wavelength Division Multiplexing (2km) Single mode fiber QSFP28 module'],
                           '100g-cr4': ['100G Direct Attach Copper (CR4 DAC) cable QSFP28 to QSFP28 Twinax','4x25G Breakout Direct Attach Copper (CR4 DAC) cable QSFP28 to 4x25G SFP+ Twinax','2x50G Breakout Direct Attach Copper (CR2 DAC) cable QSFP28 to 2x50G QSFP28 Twinax - 100G QSFP28 end','50G Direct Attach Copper (DAC) cable QSFP28 to 2xQSFP28 Twinax - 50G QSFP28 end'],
                           '100g-clr4': ['100G CLR4 Alliance Long Range (CLR - 2km) Single mode fiber QSFP28 module'],
                           '100g-swdm4': ['100G Shortwave Wavelength Division Multiplexing (SWDM4) Multimode fiber QSFP28 module'],
                           '100g-pam4-bidi': ['100G PAM4 BiDi QSFP28 module'],
                           '100g-active': ['100G Active Optical Cable (AOC) QSFP28 to QSFP28 attached fiber','4x25G Breakout Active Optical Cable (AOC) QSFP28 to 4x25G SFP+ attached fiber'],
                           '100g-dwdm2': ['100G DWDM2 - 2 wavelengths on 1550nm DWDM grid, (80km) QSFP28 module'],
                           '100g-wdm': ['100G 1550nm WDM - 4 wavelengths QSFP28 module'],
                           '40g-er4': ['40G Extended Range (ER4) Single mode fiber QSFP+ module'],
                           '4x10g-sr': ['4x10G Breakout Short Range (SR4) Multimode fiber QSFP+ module'],
                           '40g-psm4': ['40G Parallel Single Mode 4-channel (PSM4) Single mode fiber QSFP+ module'],
                           '40g-swdm4': ['40G Shortwave Wavelength Division Multiplexing (SWDM4) Single mode fiber QSFP+ module'],
                           '40g-active': ['40G Active Optical Cable (AOC) QSFP+ to QSFP+ attached fiber','4x10G Breakout Active Optical Cable (AOC) QSFP+ to 4xSFP+ attached fiber'],
                           '40g-lr4': ['40G Long Range (LR4) Single mode fiber QSFP+ module'],
                           '40g-sr4': ['40G Short Range (SR4) Multimode fiber QSFP+ module'],
                           '40g-cr4': ['40G Direct Attach Copper (CR4 DAC) cable QSFP+ to QSFP+ Twinax','4x10G Breakout Direct Attach Copper (DAC) cable QSFP+ to 4xSFP+ Twinax'],

}


ETH_SFP_CODE_MAP = {
                           '100g-active': ['25G Active Optical Cable (AOC) SFP+ to SFP+ attached fiber'],
                           '100g-sr4': ['25G Short Range (SR) fiber SFP+ module'],
                           '100g-lr4': ['25G Long Range (LR) fiber SFP+ module'],
                           '100g-active-cu': ['25G Active Copper Direct Attach Copper (DAC) cable SFP+ to SFP+ Twinax w/ amplifier and equalization'],
                           '100g-cr4': ['25G Direct Attach Copper (DAC) cable SFP+ Twinax CA-L Long Range RS FEC Required'],
                           '1000base-sx': ['1G Short Range (SX) Multimode fiber SFP module'],
                           '1000base-lx': ['1G Long Range (LX) Single mode fiber SFP module'],
                           '1000base-cx': ['1GBase-CX (balanced copper)'],
                           '1000base-t': ['1G Base-T RJ45 SFP module'],
                           '10g-t': ['10G Base-T RJ45 SFP+ module'],
                           '10g-tsr': ['10G Base-T RJ45 SFP+ module'],
                           '10g-cr': ['10G Direct Attach Copper (DAC) cable SFP+ to SFP+ Twinax'],
                           '10g-sr': ['10G Short Range (SR) Multimode fiber SFP+ module'],
                           '10g-lr': ['10G Long Range (LR - 10km) Single Mode fiber SFP+ module'],
                           '10g-lrm':['10G Long Range (LRM - 220m) Multimode fiber SFP+ module'],
                           '10g-er': ['10G Extended Range (ER - 40km) Single mode fiber SFP+ module'],
                           '25g-cr-ca-s': ['25G Direct Attach Copper (DAC) cable SFP+ Twinax CA-S Short Range BaseR FEC Required'],
                           '25g-cr-ca-n': ['25G Direct Attach Copper (DAC) cable SFP+ Twinax CA-N No FEC Required'],
                           '10g-active': ['10G Active Optical Cable (AOC) SFP+ to SFP+ attached fiber']
}


def command(command, support_file=None, meta=False):
    def decorator(func):
        def wrapper():
            return func, command, support_file, meta
        return wrapper
    return decorator


@command("/usr/lib/cumulus/bcmcmd {unit}: phy diag {sdk} dsc")
def parse_bcm_eye(raw_data):
    def lane(dat, n):
        try:
            return dat[str(n)]
        except KeyError:
            return dat["Lane{}".format(n)]

    if raw_data is None:
        return {}
    if "segment4" not in raw_data or "eye_left" not in lane(raw_data["segment4"], 0):
        return {}
    dat = raw_data["segment4"]
    mask = raw_data["lane_mask"]

    active_lanes = [i for i in range(0, 4) if 2**i & int(mask,base=16) != 0]

    return {"eye": [{"L": int(lane(dat, i)["eye_left"]),
                     "R": int(lane(dat, i)["eye_right"]),
                     "U": int(lane(dat, i)["eye_up"]),
                     "D": int(lane(dat, i)["eye_down"])} for i in active_lanes],
            "asic": "broadcom"}


@command("mlxlink -d $(mst status | grep pciconf | awk '{{print $1}}') -c -m -e --show_device --show_fec -p $(echo {port} | sed 's/swp//' | sed 's/s[0-3]//')")
def parse_mlx_eye(raw_data):
    def safe_int(val):
        try:
            return int(val)
        except ValueError:
            return 0
    if raw_data is None:
        return {}
    dat = raw_data["EYE Opening Info"]
    return {"eye": [{"grade": safe_int(dat["Physical Grade"].split(',')[i].strip()),
                    "mV": safe_int(dat["Height Eye Opening [mV]"].split(',')[i].strip())}
                    for i in range(0,len(dat["Phase  Eye Opening [psec]"].split(',')))],
            "asic": "mellanox"}


@command("ip link show {port}")
def parse_link(raw_data):
    return {"link_status": raw_data.items()[0][1]["state"]}


@command("ethtool -m {port} hex on", "ethtool.module-hex")
def parse_eeprom(raw_data):
    set_logger("cable-qualification")

    eeprom = raw_data

    if len(eeprom) == 0:
        return {}

    results = {}

    cable = SFFCable(eeprom)

    results["compliance_codes"] = cable.ethcodes_safe
    results["identifier"] = cable.identifier
    results["connector"] = cable.connector
    results["bitrate"] = cable.decoder.decode_bit_rate(eeprom) if cable.decoder is not None else None

    if "QSFP" in results["identifier"]:
        code_dict = ETH_QSFP_CODE_MAP
    else:
        code_dict = ETH_SFP_CODE_MAP

    try:
        if results["compliance_codes"][0] == "25g-cr":
            # Attempt to manually decode extended compliance code
            results["compliance_codes"] = (cable.decoder.ETH_EXTENDED_CODE_MAP.get(struct.unpack('B', 
                        eeprom[cable.decoder.ADDR_ETH_EXTENDED_CODE])[0],"25g-cr"),)
            prim_code = results["compliance_codes"][0] if results["compliance_codes"][0] != "25g-cr" else "25g-cr-ca-n"
            results["description"] = [code_dict[prim_code]]
            results["primary_code"] = prim_code
        else:
            results["description"] = [d for c in results["compliance_codes"] for d in code_dict[c]]
            results["primary_code"] = results["compliance_codes"][0]
    except Exception:
        # Invalid cable
        return {}

    results["vendor"] = {
        "oui": cable.vendor_oui,
        "pn": cable.vendor_pn,
        "name": cable.vendor_name,
        "rev": cable.vendor_rev
    }

    results["DOM"] = cable.dom

    return results


@command("lldpctl {port}", meta=True)
def parse_neighbor(raw_data):
    if 1 in raw_data and "Software Revision" in raw_data[1]:
        return {"neighbor": {
                             "os_version": raw_data[1]["Software Revision"],
                             "fw_version": raw_data[1]["Firmware Revision"],
                             "serial_no": raw_data[1].get("Serial Number", None),
                             "manufacturer": raw_data[1]["Manufacturer"],
                             "model": raw_data[1]["Model"]}
        }
    else:
        return {}
